@@ -338,7 +338,7 @@ namespace linalg
             }
         }
 
-        return std::make_tuple(eig_vals, eig_vecs);
+        return std::make_tuple(std::move(eig_vals), std::move(eig_vecs));
     }
 
     template <class E, std::enable_if_t<xtl::is_complex<typename E::value_type>::value>* = nullptr>
@@ -365,7 +365,7 @@ namespace linalg
             throw std::runtime_error("Eigenvalue calculation did not converge.");
         }
 
-        return std::make_tuple(w, VR);
+        return std::make_tuple(std::move(w), std::move(VR));
     }
 
     /**
@@ -391,7 +391,7 @@ namespace linalg
             throw std::runtime_error("Eigenvalue computation did not converge.");
         }
 
-        return std::make_tuple(w, M);
+        return std::make_tuple(std::move(w), std::move(M));
     }
 
     template <class E, std::enable_if_t<xtl::is_complex<typename E::value_type>::value>* = nullptr>
@@ -412,7 +412,7 @@ namespace linalg
             throw std::runtime_error("Eigenvalue computation did not converge.");
         }
 
-        return std::make_tuple(w, M);
+        return std::make_tuple(std::move(w), std::move(M));
     }
 
     /**
@@ -876,21 +876,11 @@ namespace linalg
 
         using common_type = std::common_type_t<typename T::value_type, typename O::value_type>;
 
-        const auto da = a.derived_cast();
-        const auto db = b.derived_cast();
-
-        XTENSOR_ASSERT(da.dimension() == 1);
-        XTENSOR_ASSERT(db.dimension() == 1);
+        XTENSOR_ASSERT(da.derived_cast().dimension() == 1);
+        XTENSOR_ASSERT(db.derived_cast().dimension() == 1);
 
         common_type result = 0;
-        if (xtl::is_complex<typename T::value_type>::value)
-        {
-            blas::dot(da, db, result);
-        }
-        else
-        {
-            blas::dotu(da, db, result);
-        }
+        blas::dot(a, b, result);
 
         return result;
     }
@@ -908,16 +898,13 @@ namespace linalg
         using common_type = std::common_type_t<typename T::value_type, typename O::value_type>;
         using return_type = xtensor<common_type, 2>;
 
-        const auto da = a.derived_cast();
-        const auto db = b.derived_cast();
+        XTENSOR_ASSERT(a.derived_cast().dimension() == 1);
+        XTENSOR_ASSERT(b.derived_cast().dimension() == 1);
 
-        XTENSOR_ASSERT(da.dimension() == 1);
-        XTENSOR_ASSERT(db.dimension() == 1);
+        typename return_type::shape_type s = {a.derived_cast().shape()[0], b.derived_cast().shape()[0]};
+        return_type result(s, common_type(0));
 
-        typename return_type::shape_type s = {da.shape()[0], db.shape()[0]};
-        return_type result(s, 0);
-
-        blas::ger(da, db, result);
+        blas::ger(a, b, result);
 
         return result;
     }
@@ -994,7 +981,7 @@ namespace linalg
             v_sign *= (LU(i, i) / abs_elem);
             result += std::log(abs_elem);
         }
-        return std::make_tuple(v_sign, result);
+        return std::make_tuple(std::move(v_sign), std::move(result));
     }
 
     /// @cond DOXYGEN_INCLUDE_SFINAE
@@ -1033,7 +1020,7 @@ namespace linalg
         }
 
         value_type v_sign = (sign % 2) ? -1 : 1;
-        return std::make_tuple(v_sign, result);
+        return std::make_tuple(std::move(v_sign), std::move(result));
     }
     /// @endcond
 
@@ -1062,6 +1049,21 @@ namespace linalg
         }
     }
 
+    namespace xblas_detail
+    {
+        template <class T>
+        inline void triu_inplace(T& R)
+        {
+            for (std::size_t i = 0; i < R.shape()[0]; ++i)
+            {
+                for (std::size_t j = 0; j < i && j < R.shape()[1]; j++)
+                {
+                    R(i, j) = 0;
+                }
+            }
+        }
+    }
+
     /// Select the mode for the qr decomposition ``K = min(M, K)``
     enum class qrmode {
         reduced,  ///< return Q, R with dimensions (M, K), (K, N) (default)
@@ -1079,8 +1081,7 @@ namespace linalg
     auto qr(const xexpression<T>& A, qrmode mode = qrmode::reduced)
     {
         using value_type = typename T::value_type;
-        using xtype = xtensor<value_type, 2, layout_type::column_major>;
-        using result_xtype = xtensor<value_type, 2>;
+        using xtype = xarray<value_type, layout_type::column_major>;
 
         xtype R = A.derived_cast();
 
@@ -1088,8 +1089,7 @@ namespace linalg
         std::size_t N = R.shape()[1];
         std::size_t K = std::min(M, N);
 
-        std::array<std::size_t, 2> tau_shp = {K, 1};
-        xtype tau(tau_shp);
+        auto tau = xarray<value_type, layout_type::column_major>::from_shape({K});
         int info = lapack::geqrf(R, tau);
 
         if (info != 0)
@@ -1097,43 +1097,44 @@ namespace linalg
             throw std::runtime_error("QR decomposition failed.");
         }
 
-        xtype Q;
+        // explicitly set shape/size == 0!
+        auto Q = xtype::from_shape({0});
+
+        if (mode == qrmode::r)
+        {
+            R = xt::view(R, range(0, K), all());
+            xblas_detail::triu_inplace(R);
+            return std::make_tuple(std::move(Q), std::move(R));
+        }
 
         if (mode == qrmode::raw)
         {
-            return std::make_tuple(R, tau);
+            R = transpose(R);
+            return std::make_tuple(std::move(R), std::move(tau));
         }
 
-        if (mode == qrmode::reduced)
-        {
-            Q = R;
-            detail::call_gqr(Q, tau, static_cast<blas_index_t>(K));
-            auto vR = view(R, range(std::size_t(0), K), all());
-            R = vR;
-        }
-        if (mode == qrmode::complete)
+        blas_index_t mc;
+
+        if (mode == qrmode::complete && M > N)
         {
+            mc = (blas_index_t) M;
             Q.resize({M, M});
-            // TODO replace with assignment to view
-            for (std::size_t i = 0; i < R.shape()[0]; ++i)
-            {
-                for (std::size_t j = 0; j < R.shape()[1]; ++j)
-                {
-                    Q(i, j) = R(i, j);
-                }
-            }
-            detail::call_gqr(Q, tau, static_cast<blas_index_t>(M));
         }
-
-        for (std::size_t i = 0; i < R.shape()[0]; ++i)
+        else
         {
-            for (std::size_t j = 0; j < i && j < R.shape()[1]; j++)
-            {
-                R(i, j) = 0;
-            }
+            mc = (blas_index_t) K;
+            Q.resize({M, N});
         }
 
-        return std::make_tuple(Q, R);
+        xt::view(Q, all(), range(0, N)) = R;
+        detail::call_gqr(Q, tau, mc);
+
+        Q = xt::view(Q, all(), range(0, mc));
+        R = xt::view(R, range(0, mc), all());
+
+        xblas_detail::triu_inplace(R);
+
+        return std::make_tuple(std::move(Q), std::move(R));
     }
 
     /**
@@ -1192,7 +1193,7 @@ namespace linalg
             throw std::runtime_error("SVD decomposition failed.");
         }
 
-        return std::make_tuple(std::get<1>(result), std::get<2>(result), std::get<3>(result));
+        return std::make_tuple(std::move(std::get<1>(result)), std::move(std::get<2>(result)), std::move(std::get<3>(result)));
     }
 
     /**
@@ -1208,9 +1209,9 @@ namespace linalg
 
         auto gesdd_res = svd(M, false, true);
 
-        auto u = std::get<0>(gesdd_res);
-        auto s = std::get<1>(gesdd_res);
-        auto vt = std::get<2>(gesdd_res);
+        auto u = std::move(std::get<0>(gesdd_res));
+        auto s = std::move(std::get<1>(gesdd_res));
+        auto vt = std::move(std::get<2>(gesdd_res));
 
         using real_value_type = typename decltype(s)::value_type;
         real_value_type cutoff = static_cast<real_value_type>(rcond) * (*std::max_element(s.begin(), s.end()));
@@ -1420,23 +1421,45 @@ namespace linalg
         using underlying_value_type = xtl::complex_value_type_t<typename T::value_type>;
 
         xtensor<value_type, 2, layout_type::column_major> dA = A.derived_cast();
-        xtensor<value_type, 2, layout_type::column_major> db;
-
-        const auto& db_t = b.derived_cast();
-        if (db_t.dimension() == 1)
-        {
-            std::size_t sz = db_t.shape()[0];
-            db.resize({sz, 1});
-            std::copy(db_t.storage().begin(), db_t.storage().end(), db.storage().begin());
-        }
-        else
-        {
-            db = db_t;
-        }
 
         std::size_t M = dA.shape()[0];
         std::size_t N = dA.shape()[1];
 
+        auto& b_ref = b.derived_cast();
+
+        if (dA.dimension() != 2)
+        {
+            throw std::runtime_error("Expected 2D expression for A");
+        }
+
+        if (!(b_ref.dimension() <= 2))
+        {
+            throw std::runtime_error("Expected 1- or 2D expression for A.");
+        }
+
+        if (b_ref.shape()[0] != M)
+        {
+            throw std::runtime_error("Shape of 'b' for lstsq does not match.");
+        }
+
+        // find number of rhs
+        std::size_t nrhs = (b_ref.dimension() == 1) ? 1 : b_ref.shape()[1];
+
+        // as the dgelsd docs say, on entry it's M-by-nrhs, then result N-by-nrhs
+        // that is why we need to allocate *MORE* space than just b here for M > N
+        auto db = xarray<value_type, layout_type::column_major>::from_shape({ std::max(M, N), nrhs });
+
+        bool is_1d = false;
+        if (b_ref.dimension() == 1)
+        {
+            is_1d = true;
+            xt::view(db, range(0, M), xt::all()) = xt::view(b_ref, xt::all(), xt::newaxis());
+        }
+        else
+        {
+            xt::view(db, range(0, M), xt::all()) = b_ref;
+        }
+
         auto s = xtensor<underlying_value_type, 1, layout_type::column_major>::from_shape({ std::min(M, N) });
 
         blas_index_t rank;
@@ -1460,10 +1483,14 @@ namespace linalg
             }
         }
 
-        auto vdb = view(db, range(std::size_t(0), N));
+        auto vdb = view(db, range(std::size_t(0), N), xt::all());
         db = vdb;
+        if (is_1d)
+        {
+            db = xt::squeeze(db);
+        }
 
-        return std::make_tuple(db, residuals, rank, s);
+        return std::make_tuple(std::move(db), std::move(residuals), std::move(rank), std::move(s));
     }
 
     /**
@@ -1685,4 +1712,4 @@ namespace linalg
     }
 }
 }
-#endif
+#endif
\ No newline at end of file
